import std.stdio;
import std.string;
import std.utf;
import std.conv;
import core.stdc.ctype;

class StringDB {
  ubyte[][] array ;
  size_t    array_length ;
  int       maxclm=0 ;
  int       maxline=0 ;
  wstring   caption ;
  int       fmt_type=0 ;
  int       fmt_size=1 ;
  const int blocksize=65536 ;
  wstring[] delegate(int,int) fmt_tbl[int] ;

  this(){
    fmt_tbl[0] = &formatBigEndian ;
    fmt_tbl[1] = &formatLittleEndian ;
  }

  ~this(){
    for( int i=0 ; i<array.length ; i++ ){
      delete array[i] ;
    }
    array.length=0 ;
  }

  int getMaxColumns(){
    return(maxclm) ;
  }

  int getMaxLines(){
    return(maxline+1) ;
  }

  wstring[] getStringDB(int pos){
    return(fmt_tbl[fmt_type](fmt_size,pos)) ;
  }

  wstring getCaptionString(){
    return(caption) ;
  }

  int setFormat(int type, int size){
    fmt_type=type ;
    fmt_size=size ;
    fmt_tbl[fmt_type](fmt_size,0) ;
    return(0) ;
  }

  int[] getFormat(){
    int[] ret = new int[2] ;
    ret[0]=fmt_type ;
    ret[1]=fmt_size ;
    return(ret) ;
  }

  private wstring[] formatBigEndian(int size, int pos){
    wstring[] warray ;
    warray.length=1 ;
    if( pos<maxline ){
      size_t ptr=pos*16 ;
      foreach( i, ref l ; warray){
        string line=format("%08x : ",ptr) ;
        for( int j=0 ; j<16 ; j++ ){
          if( ptr+j >= array_length ){
            line ~= "  " ;
          }else{
            line ~= format("%02x",array[(ptr+j)/blocksize][(ptr+j)%blocksize]) ;
          }
          if( ((j+1)%size)==0 ) line ~= " " ;
        }
        line ~= " " ;
        for( int j=0 ; j<16 ; j++ ){
          if( ptr+j >= array_length ) break ;
          char[1] c ; c[0] = array[(ptr+j)/blocksize][(ptr+j)%blocksize] ;
          line ~= isprint(c[0])!=0 ?  c : "." ;
        }
        l=toUTF16(line) ;
        ptr+=16 ;
      }
    }

    switch(size){
    case 1: caption="address  : +0 +1 +2 +3 +4 +5 +6 +7 +8 +9 +a +b +c +d +e +f  0123456789abcdef" ; break ;
    case 2: caption="address  : +0+1 +2+3 +4+5 +6+7 +8+9 +a+b +c+d +e+f  0123456789abcdef"         ; break ;
    case 4: caption="address  : +0+1+2+3 +4+5+6+7 +8+9+a+b +c+d+e+f  0123456789abcdef"             ; break ;
    case 8: caption="address  : +0+1+2+3+4+5+6+7 +8+9+a+b+c+d+e+f  0123456789abcdef"               ; break ;
    default : caption="" ; break ;
    }
    maxclm=cast(int)(caption.length) ;

    return(warray) ;
  }

  private wstring[] formatLittleEndian(int size, int pos){
    wstring[] warray ;
    warray.length=1 ;
    if( pos<maxline ){
      size_t ptr=(maxline-pos-1)*16 ;
      foreach( i, ref l ; warray){
        string line=format("%08x : ",ptr) ;
        for( int j=15 ; j>=0 ; j-- ){
          if( ptr+j >= array_length ){
            line ~= "  " ;
          }else{
            line ~= format("%02x",array[(ptr+j)/blocksize][(ptr+j)%blocksize]) ;
          }
          if( (j%size)==0 ) line ~= " " ;
        }
        line ~= " " ;
        for( int j=15 ; j>=0 ; j-- ){
          char[1] c ;
          if( ptr+j >= array_length ){
            c[0]=' ' ;
          }else{
            c[0] = array[(ptr+j)/blocksize][(ptr+j)%blocksize] ;
          }
          line ~= isprint(c[0])!=0 ?  c : "." ;
        }
        l=toUTF16(line) ;
        ptr-=16 ;
      }
    }

    switch(size){
    case 1: caption="address  : +f +e +d +c +b +a +9 +8 +7 +6 +5 +4 +3 +2 +1 +0  fedcba9876543210" ; break ;
    case 2: caption="address  : +f+e +d+c +b+a +9+8 +7+6 +5+4 +3+2 +1+0  fedcba9876543210"         ; break ;
    case 4: caption="address  : +f+e+d+c +b+a+9+8 +7+6+5+4 +3+2+1+0  fedcba9876543210"             ; break ;
    case 8: caption="address  : +f+e+d+c+b+a+9+8 +7+6+5+4+3+2+1+0  fedcba9876543210"               ; break ;
    default : caption="" ; break ;
    }
    maxclm=cast(int)(caption.length) ;

    return(warray) ;
  }

  int readfile(string fname){
    array_length=0 ;
    for( int i=0 ; i<array.length ; i++ ){
      delete array[i] ;
    }
    array.length=0 ;
    core.memory.GC.minimize() ;
    File infile ;
    if( fname != "-" ){
      infile = File(fname) ;
    }else{
      infile = stdin ;
    }
    while(1){
      ubyte[] rdbuf ;
      size_t  blksize ;
      rdbuf = infile.rawRead(new ubyte[blocksize]) ;
      array ~= rdbuf ;
      array_length += rdbuf.length ;
      if( infile.eof() == true ) break ;
    }
    if( fname != "-" ) infile.detach() ;
    maxline = cast(int)(array_length/16 + (array_length%16>0 ? 1 : 0 )) ;

    core.memory.GC.minimize() ;
    return(0) ;
  }

}
