import win32.windows;
import win32.windef;
import win32.commctrl;
import win32frame.WindowObject;
import std.stdio;
import std.string;
import std.utf;

import FontMan;
import StringDB;

COLORREF RGBA(uint r, uint g, uint b, uint a) {
  return cast(COLORREF)(r | (g << 8) | (b << 16) | (a<<24));
}

const LBUTTONDOWN   = (0x0001) ;
const LBUTTONUP     = (0x0002) ;
const LBUTTONDBLCLK = (0x0004) ;
const RBUTTONDOWN   = (0x0010) ;
const RBUTTONUP     = (0x0020) ;
const RBUTTONDBLCLK = (0x0040) ;
const ALT_KEY       = (0x0100) ;
const CNTL_KEY      = (0x0200) ;
const SHIFT_KEY     = (0x0400) ;

class TextView : WindowObject {
  HFONT      txtfont ;
  FontMan    fontman ;
  SCROLLINFO rollinfo_H ;
  SCROLLINFO rollinfo_V ;

  StringDB strdb ;

  int  max_width =0  ;
  int  max_height=0  ;
  int  curline   =0  ;
  int  curcolm   =0  ;

  int fontheight ;
  int fontwidth  ;


  this(string name, string ClassName) {
    winc.style          = 0; //CS_HREDRAW | CS_VREDRAW;
    winc.cbClsExtra     = 0;
    winc.cbWndExtra     = 0;
    winc.lpszMenuName   = null;
    winc.hCursor        = LoadCursor(null , cast(wchar*)(IDC_ARROW));
    winc.lpszClassName  = toUTFz!(wchar*)(ClassName) ;
    super(name) ;
    this.registwinc() ;
  }

  override LRESULT proc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    switch (msg) {
    case WM_DESTROY:
      break ;

    case WM_CREATE:
      fontman.init(-12,FIXED_PITCH,"ＭＳ ゴシック") ;

      fontheight=fontman.lf.lfHeight<0 ? -fontman.lf.lfHeight : fontman.lf.lfHeight ;
      fontwidth =fontheight/2+ fontheight%2;

      rollinfo_H.cbSize = SCROLLINFO.sizeof ;
      rollinfo_V.cbSize = SCROLLINFO.sizeof ;

      rollinfo_H.fMask  = SIF_RANGE | SIF_POS | SIF_PAGE | SIF_DISABLENOSCROLL ;
      rollinfo_H.nPos   = 0 ;
      rollinfo_H.nPage  = 2 ;
      rollinfo_H.nMin   = 0 ;
      rollinfo_H.nMax   = 1 ;
      SetScrollInfo(hwnd, SB_HORZ, &rollinfo_H, true) ;

      rollinfo_V.cbSize = SCROLLINFO.sizeof ;
      rollinfo_V.fMask  = SIF_RANGE | SIF_POS | SIF_PAGE | SIF_DISABLENOSCROLL ;
      rollinfo_V.nPos   = 0 ;
      rollinfo_V.nPage  = 2 ;
      rollinfo_V.nMin   = 0 ;
      rollinfo_V.nMax   = 1 ;
      SetScrollInfo(hwnd, SB_VERT, &rollinfo_V, true) ;

      break ;

    case WM_HSCROLL:
      int changed=1 ;
      rollinfo_H.fMask  = SIF_POS ;
      switch( LOWORD(wp) ){
      case SB_LINELEFT :
        rollinfo_H.nPos -= 1 ;
        if( rollinfo_H.nPos < 0 ) rollinfo_H.nPos=0 ;
        break ;
      case SB_LINERIGHT :
        rollinfo_H.nPos += 1 ;
        if( rollinfo_H.nPos > rollinfo_H.nMax ) rollinfo_H.nPos=rollinfo_H.nMax ;
        break ;
      case SB_PAGELEFT :
        rollinfo_H.nPos -= rollinfo_H.nPage/2 ;
        if( rollinfo_H.nPos < 0 ) rollinfo_H.nPos=0 ;
        break ;
      case SB_PAGERIGHT :
        rollinfo_H.nPos += rollinfo_H.nPage/2 ;
        if( rollinfo_H.nPos > rollinfo_H.nMax ) rollinfo_H.nPos=rollinfo_H.nMax ;
        break ;
      case SB_THUMBTRACK :
        {
          SCROLLINFO  tmp_scrinfo ;
          tmp_scrinfo.cbSize = SCROLLINFO.sizeof ; tmp_scrinfo.fMask = SIF_TRACKPOS ;
          GetScrollInfo( hwnd, SB_HORZ, &tmp_scrinfo) ;
          rollinfo_H.nPos = tmp_scrinfo.nTrackPos ;
        }
        break ;

      default:
        changed=0 ;
        break ;
      }

      if( changed!=0 ){
        SetScrollInfo( hwnd, SB_HORZ, &rollinfo_H, true) ;
        InvalidateRect(hwnd , null , true);
        curcolm = rollinfo_H.nPos ;
        sendmsgtoparent( WM_COMMAND, cast(int)(getHMENU()), cast(LPARAM)(SB_HORZ) ) ;
      }
      break ;

    case WM_VSCROLL:
      int changed=1 ;
      rollinfo_V.fMask  = SIF_POS ;
      switch( LOWORD(wp) ){
      case SB_LINEUP :
        rollinfo_V.nPos -= 1 ;
        if( rollinfo_V.nPos < 0 ) rollinfo_V.nPos=0 ;
        break ;
      case SB_LINEDOWN :
        rollinfo_V.nPos += 1 ;
        if( rollinfo_V.nPos > rollinfo_V.nMax ) rollinfo_V.nPos=rollinfo_V.nMax ;
        break ;
      case SB_PAGEUP :
        rollinfo_V.nPos -= rollinfo_V.nPage/2 ;
        if( rollinfo_V.nPos < 0 ) rollinfo_V.nPos=0 ;
        break ;
      case SB_PAGEDOWN :
        rollinfo_V.nPos += rollinfo_V.nPage/2 ;
        if( rollinfo_V.nPos > rollinfo_V.nMax ) rollinfo_V.nPos=rollinfo_V.nMax ;
        break ;
      case SB_THUMBTRACK :
        {
          SCROLLINFO  tmp_scrinfo ;
          tmp_scrinfo.cbSize = SCROLLINFO.sizeof ; tmp_scrinfo.fMask = SIF_TRACKPOS ;
          GetScrollInfo(hwnd, SB_VERT, &tmp_scrinfo) ;
          rollinfo_V.nPos = tmp_scrinfo.nTrackPos ;
        }
        break ;

      default:
        changed=0 ;
        break ;
      }
      if( changed!=0 ){
        SetScrollInfo( hwnd, SB_VERT, &rollinfo_V, true) ;
        InvalidateRect(hwnd , null , true);
        curline = rollinfo_V.nPos ;
        sendmsgtoparent( WM_COMMAND, cast(WPARAM)(getHMENU()), cast(LPARAM)(SB_VERT) ) ;
      }
      break ;

    case WM_PAINT:
      SCROLLINFO  srinfo_H ;
      SCROLLINFO  srinfo_V ;
      srinfo_H.cbSize = SCROLLINFO.sizeof ; srinfo_H.fMask = SIF_POS ;
      srinfo_V.cbSize = SCROLLINFO.sizeof ; srinfo_V.fMask = SIF_POS ;
      GetScrollInfo( hwnd, SB_HORZ, &srinfo_H) ;
      GetScrollInfo( hwnd, SB_VERT, &srinfo_V) ;

      PAINTSTRUCT ps;
      HDC         hdc;
      HPEN        PBG,PHD ;
      HBRUSH      BBG,BHD ;
      PBG  = CreatePen(PS_SOLID, 1, RGBA(255,255,255,  0) ) ;
      BBG  = CreateSolidBrush( RGBA(255,255,255,0) ) ;
      PHD  = CreatePen(PS_SOLID, 1, RGBA(192, 192, 192,  0) ) ;
      BHD  = CreateSolidBrush( RGBA(192,192,192,0) ) ;

      hdc = BeginPaint(hwnd, &ps);
      //SetBkMode(hdc,TRANSPARENT) ;
      txtfont = fontman.createHFONT() ;
      SelectObject(hdc , txtfont);

      SelectObject(hdc, PBG) ;
      SelectObject(hdc, BBG) ;
      SetBkColor(hdc,RGBA(255,255,255,0)) ;
      int drawlines = winH/fontheight+ (winH%fontheight!=0) ;

      for( int i=0 ; i<drawlines ; i++ ){
        int ptr=srinfo_V.nPos+(i-1) ;
        if( strdb !is null ){
          wstring[] array = strdb.getStringDB(ptr) ;
          if( i==0 ){
            SelectObject(hdc, PBG) ;
            SelectObject(hdc, BBG) ;
            SetBkColor(hdc,RGBA(255,255,255,0)) ;
            SIZE size ;
            wstring caption=strdb.getCaptionString() ;
            GetTextExtentPoint32(hdc, cast(wchar*)(caption), cast(int)(caption.length), &size) ;
            TextOut(hdc, 0-srinfo_H.nPos*fontwidth, i*fontheight, cast(wchar*)(caption) , cast(int)(caption.length));
            Rectangle(hdc, size.cx-(srinfo_H.nPos*fontwidth), i*fontheight, winW, (i+1)*fontheight) ;
            continue ;
          }
          if( array.length>0 ){
            SelectObject(hdc, PBG) ;
            SelectObject(hdc, BBG) ;
            SetBkColor(hdc,RGBA(255,255,255,0)) ;
            SIZE size ;
            GetTextExtentPoint32(hdc, cast(wchar*)(array[0]), cast(int)(array[0].length), &size) ;
            TextOut(hdc, 0-srinfo_H.nPos*fontwidth, i*fontheight, cast(wchar*)(array[0]) , cast(int)(array[0].length));
            Rectangle(hdc, size.cx-(srinfo_H.nPos*fontwidth), i*fontheight, winW, (i+1)*fontheight) ;
            continue ;
          }
        }
        Rectangle(hdc, 0, i*fontheight, winW, (i+1)*fontheight) ;
      }

      EndPaint(hwnd, &ps);
      DeleteObject(PBG ) ;
      DeleteObject(BBG ) ;
      DeleteObject(PHD ) ;
      DeleteObject(BHD ) ;
      DeleteObject(txtfont);
      break ;

    case WM_SIZE:
      if( strdb !is null ){
        int winw   = LOWORD(lp)/fontwidth + ( LOWORD(lp)%fontwidth !=0 ) ;
        int winh   = HIWORD(lp)/fontheight+ ( HIWORD(lp)%fontheight!=0 ) ;
        int width  = strdb.getMaxColumns() ;
        int height = strdb.getMaxLines()   ;
        rollinfo_H.fMask  = SIF_RANGE | SIF_PAGE | SIF_DISABLENOSCROLL ;
        rollinfo_H.nPage  = winw     ;
        rollinfo_H.nMin   = 0        ;
        rollinfo_H.nMax   = width    ;
        SetScrollInfo(hwnd, SB_HORZ, &rollinfo_H, true) ;

        rollinfo_V.cbSize = SCROLLINFO.sizeof ;
        rollinfo_V.fMask  = SIF_RANGE | SIF_PAGE | SIF_DISABLENOSCROLL ;
        rollinfo_V.nPage  = winh     ;
        rollinfo_V.nMin   = 0        ;
        rollinfo_V.nMax   = height   ;
        SetScrollInfo(hwnd, SB_VERT, &rollinfo_V, true) ;
      }
      RedrawWindow( hwnd, null, null, RDW_INVALIDATE|RDW_UPDATENOW) ;
      break;

    case WM_MOUSEWHEEL:
      int dlt=cast(int)(wp)>>16 ;
      if( (LOWORD(wp)&MK_SHIFT)==0 ){
        int dispH = winH/fontheight+ (winH%fontheight!=0) ;
        rollinfo_V.fMask  = SIF_POS ;
        rollinfo_V.nPos -=dlt/10 ;
        if( rollinfo_V.nPos <  0                       ) rollinfo_V.nPos=0 ;
        if( rollinfo_V.nPos >= (rollinfo_V.nMax-dispH) ) rollinfo_V.nPos=rollinfo_V.nMax-dispH+1 ;
        InvalidateRect(hwnd , null , true);
        SetScrollInfo(hwnd, SB_VERT, &rollinfo_V, true) ;
        curline = rollinfo_V.nPos ;
        sendmsgtoparent( WM_COMMAND, cast(WPARAM)(getHMENU()), cast(LPARAM)(SB_VERT) ) ;
      }else{
        int dispW = winW/fontwidth + (winW%fontwidth!=0) ;
        rollinfo_H.fMask  = SIF_POS ;
        rollinfo_H.nPos -=dlt/10 ;
        if( rollinfo_H.nPos <  0                       ) rollinfo_H.nPos=0 ;
        if( rollinfo_H.nPos >= (rollinfo_H.nMax-dispW) ) rollinfo_H.nPos=rollinfo_H.nMax-dispW+1 ;
        InvalidateRect(hwnd , null , true);
        SetScrollInfo(hwnd, SB_HORZ, &rollinfo_H, true) ;
        curcolm = rollinfo_H.nPos ;
        sendmsgtoparent( WM_COMMAND, cast(WPARAM)(getHMENU()), cast(LPARAM)(SB_HORZ) ) ;
      }
      break ;

    case WM_LBUTTONDOWN:
    case WM_LBUTTONUP:
    case WM_LBUTTONDBLCLK:
    case WM_RBUTTONDOWN:
    case WM_RBUTTONUP:
    case WM_RBUTTONDBLCLK:
    case WM_MOUSEMOVE:
      POINT point;
      SCROLLINFO srinfo_H ;
      SCROLLINFO srinfo_V ;
      GetCursorPos(&point);
      ScreenToClient( hwnd, &point );
      srinfo_H.cbSize = SCROLLINFO.sizeof ; srinfo_H.fMask = SIF_POS ;
      srinfo_V.cbSize = SCROLLINFO.sizeof ; srinfo_V.fMask = SIF_POS ;
      GetScrollInfo( hwnd, SB_HORZ, &srinfo_H) ;
      GetScrollInfo( hwnd, SB_VERT, &srinfo_V) ;
      break ;

    default :
      return( DefWindowProc(hwnd , msg , wp , lp) );
    }
    return(0);
  }

  int setViewPos(int fnBar, int nPos){
    switch( fnBar ){
    case SB_HORZ :
      rollinfo_H.fMask = SIF_POS ;
      rollinfo_H.nPos  = nPos ;
      SetScrollInfo(getHWND(), SB_HORZ, &rollinfo_H, true) ;
      break ;
    case SB_VERT :
      rollinfo_V.fMask = SIF_POS ;
      rollinfo_V.nPos  = nPos ;
      SetScrollInfo(getHWND(), SB_VERT, &rollinfo_V, true) ;
      break ;
    default :
      break ;
    }
    RedrawWindow( getHWND(), null, null, RDW_INVALIDATE|RDW_UPDATENOW) ;
    return(0) ;
  }

  int getCurLine(){
    return(curline) ;
  }

  int getCurColumn(){
    return(curcolm) ;
  }

  int setStringDB(StringDB strdb){
    this.strdb = strdb ;
    reDraw() ;
    return(0) ;
  }

  int reDraw(){
    sendmsg(WM_SIZE, 0, (winH<<16 | winW) );
    return(0) ;
  }

  int changeFont(LOGFONT lf){
    fontman.setLOGFONT(lf) ;
    fontheight=fontman.lf.lfHeight<0 ? -fontman.lf.lfHeight : fontman.lf.lfHeight ;
    fontwidth =fontheight/2+ fontheight%2;

    reDraw() ;
    return(0) ;
  }
}
