import win32.windows;
import win32.windef;
import win32.commctrl;
import win32frame.Frame;
import win32frame.WindowMenu;
import win32frame.StatusBar;
import win32frame.MainWindow ;
import win32frame.FileDialog ;
import win32frame.MsgBox ;
import win32frame.FontSelDialog ;

import MainPanel ;

import std.stdio;
import std.string;
import std.utf;

enum FILE{
  ROOT  =0x100,
  OPEN  =0x101,
  CLOSE =0x103,
  EXIT  =0x104
} ;
enum EDIT{
  ROOT  =0x200,
} ;
enum VIEW{
  ROOT  =0x300,
  FONT  =0x301,
  BIG1  =0x311,
  BIG2  =0x312,
  BIG4  =0x313,
  BIG8  =0x314,
  LTL1  =0x321,
  LTL2  =0x322,
  LTL4  =0x323,
  LTL8  =0x324
} ;
enum HELP{
  ROOT    =0x400,
  VERSION =0x401
} ;

enum CMD{
  OPENDLG =0xa00,
} ;

class myMainWindow : MainWindow {
  WindowMenu   menuroot  ;
  StatusBar    statusbar ;
  MainPanel    mainpanel ;

  HFONT        hFont;

  string       ftp_host   ;
  string       ftp_user   ;
  string       ftp_passwd ;
  string       ftp_fname  ;

  string       appname="BinaryDump" ;

  string[]     args ;

  this(string name, string ClassName) {
    winc.style          = 0; //CS_HREDRAW | CS_VREDRAW;
    winc.cbClsExtra     = 0;
    winc.cbWndExtra     = 0;
    winc.hIcon          = LoadIcon(null , cast(wchar*)(IDI_APPLICATION));
    winc.hCursor        = LoadCursor(null , cast(wchar*)(IDC_ARROW));
    winc.hbrBackground  = CreateSolidBrush(GetSysColor(COLOR_BTNFACE)) ;
    winc.lpszMenuName   = null;
    winc.lpszClassName  = toUTFz!(wchar*)(ClassName) ;
    super(name,ClassName) ;
  }

  override LRESULT proc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    switch (msg) {
    case WM_DESTROY:
      DeleteObject(hFont);
      PostQuitMessage(0);
      break ;

    case WM_CREATE:
      menuroot = new WindowMenu() ;
      menuroot.create() ;
      menuroot.addMenu(         MFT_STRING, "File"     , FILE.ROOT ) ;
      menuroot.addItem( "File", MFT_STRING, "Open"     , FILE.OPEN ) ;
      menuroot.addItem( "File", MFT_STRING, "Exit"     , FILE.EXIT ) ;
      menuroot.addMenu(         MFT_STRING, "Edit"     , EDIT.ROOT ) ;
      menuroot.addMenu(         MFT_STRING, "View"     , VIEW.ROOT ) ;
      menuroot.addItem( "View", MFT_STRING, "Font"     , VIEW.FONT ) ;
      menuroot.addItem( "View", MFT_STRING|MFT_RADIOCHECK, "Big-1byte",    VIEW.BIG1 ) ;
      menuroot.addItem( "View", MFT_STRING|MFT_RADIOCHECK, "Big-2byte",    VIEW.BIG2 ) ;
      menuroot.addItem( "View", MFT_STRING|MFT_RADIOCHECK, "Big-4byte",    VIEW.BIG4 ) ;
      menuroot.addItem( "View", MFT_STRING|MFT_RADIOCHECK, "Big-8byte",    VIEW.BIG8 ) ;
      menuroot.addItem( "View", MFT_STRING|MFT_RADIOCHECK, "Little-1byte", VIEW.LTL1 ) ;
      menuroot.addItem( "View", MFT_STRING|MFT_RADIOCHECK, "Little-2byte", VIEW.LTL2 ) ;
      menuroot.addItem( "View", MFT_STRING|MFT_RADIOCHECK, "Little-4byte", VIEW.LTL4 ) ;
      menuroot.addItem( "View", MFT_STRING|MFT_RADIOCHECK, "Little-8byte", VIEW.LTL8 ) ;
      menuroot.addMenu(         MFT_STRING, "Help"     , HELP.ROOT ) ;
      menuroot.addItem( "Help", MFT_STRING, "Version"  , HELP.VERSION) ;

      setMenu(menuroot) ;

      statusbar = new StatusBar() ;
      statusbar.addStyle(CCS_BOTTOM | SBARS_SIZEGRIP) ;
      addStatusBar(statusbar, "") ;

      hFont = CreateFont(
                         -14 , 0 , 0 , 0 , FW_REGULAR , FALSE , FALSE , FALSE ,
                         DEFAULT_CHARSET , OUT_DEFAULT_PRECIS ,
                         CLIP_DEFAULT_PRECIS , DEFAULT_QUALITY ,
                         VARIABLE_PITCH, null
                         );

      DragAcceptFiles(hwnd,true) ;

      mainpanel = new MainPanel("MainPanel","MainPanelTest") ;
      mainpanel.setPosition(0,0,200,200) ;
      mainpanel.setStyle( WS_CHILD |WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_VISIBLE ) ;
      addObject( mainpanel, null, true, true ) ;

      ftp_host   = "" ;
      ftp_user   = "" ;
      ftp_passwd = "" ;
      ftp_fname  = "" ;

      mainpanel.setFormat(0,1) ;
      if( args.length>1 ){
        try{
          mainpanel.readfile(args[1]) ;
          this.setWindowText(appname ~ " " ~ args[1]) ;
        }catch(Exception e){
          auto msgbox = new MsgBox(null, e.msg, "FileOpenError", MB_OK ) ;
          msgbox.create() ;
        }
      }
      break ;

    case WM_INITMENU:
      int[] dspfmt = mainpanel.getFormat() ;
      if( !(dspfmt is null) ){
        setDisplayFormatMenu() ;
      }
      break ;

    case WM_DROPFILES:
      int    filenum = DragQueryFile( cast(HDROP)(wp), -1, null, 0) ;
      wchar  wfname[] ;
      string fnamelist[] ;

      fnamelist.length=filenum ;
      for( int i=0 ; i<filenum ; i++ ){
        int namelen=DragQueryFile( cast(HDROP)(wp), i, null, 0) ;
        wfname.length=namelen+1 ;
        DragQueryFile( cast(HDROP)(wp), i, wfname.ptr, namelen+1) ;
        fnamelist[i]=std.utf.toUTF8(wfname.idup) ;
      }
      DragFinish( cast(HDROP)(wp) ) ;
      mainpanel.readfile(fnamelist[0]) ;
      this.setWindowText(appname ~ " " ~ fnamelist[0]) ;
      break ;

    case WM_COMMAND:
      switch( LOWORD(wp) ){
      case FILE.OPEN :
        FileDialog fdlog = new FileDialog() ;
        string     fname = fdlog.getLoadFileName("ALL {*.*}\0*.*\0",hwnd,OFN_HIDEREADONLY) ;
        if( fname !is null ){
          mainpanel.readfile(fname) ;
          this.setWindowText(appname ~ " " ~ fname) ;
        }
        break ;

      case FILE.EXIT :
        sendmsg(WM_CLOSE, 0,0) ;
        break ;

      case VIEW.FONT :
        LOGFONT    lf ;
        CHOOSEFONT cf ;
        lf = mainpanel.getViewLOGFONT() ;
        cf.lpLogFont = &lf ;
        cf.hwndOwner = hwnd ;
        cf.Flags     = CF_FIXEDPITCHONLY | CF_SCREENFONTS | CF_NOVERTFONTS | CF_INITTOLOGFONTSTRUCT | CF_SELECTSCRIPT ;
        auto fontsel = new FontSelDialog() ;
        disable() ;
        if( fontsel.chooseFont(&cf)==true ){
          mainpanel.changeFont(lf) ;
        }
        enable() ;
        break ;

      case VIEW.BIG1 : mainpanel.setFormat(0,1) ; setDisplayFormatMenu() ; break ;
      case VIEW.BIG2 : mainpanel.setFormat(0,2) ; setDisplayFormatMenu() ; break ;
      case VIEW.BIG4 : mainpanel.setFormat(0,4) ; setDisplayFormatMenu() ; break ;
      case VIEW.BIG8 : mainpanel.setFormat(0,8) ; setDisplayFormatMenu() ; break ;
      case VIEW.LTL1 : mainpanel.setFormat(1,1) ; setDisplayFormatMenu() ; break ;
      case VIEW.LTL2 : mainpanel.setFormat(1,2) ; setDisplayFormatMenu() ; break ;
      case VIEW.LTL4 : mainpanel.setFormat(1,4) ; setDisplayFormatMenu() ; break ;
      case VIEW.LTL8 : mainpanel.setFormat(1,8) ; setDisplayFormatMenu() ; break ;

      case HELP.VERSION :
        auto msgbox = new MsgBox(hwnd, appname ~ "   Ver0.01", "Version", MB_OK ) ;
        msgbox.create() ;
        break ;

      default :
        break ;
      }
      break;

    default :
      return( DefWindowProc(hwnd , msg , wp , lp) );
    }
    return(0);
  }

  void setDisplayFormatMenu(){
    int[] dspfmt = mainpanel.getFormat() ;
    if( !(dspfmt is null) ){
      int type = dspfmt[0] ;
      int size = dspfmt[1] ;
      menuroot.checkItem( "View", VIEW.BIG1, false) ;
      menuroot.checkItem( "View", VIEW.BIG2, false) ;
      menuroot.checkItem( "View", VIEW.BIG4, false) ;
      menuroot.checkItem( "View", VIEW.BIG8, false) ;
      menuroot.checkItem( "View", VIEW.LTL1, false) ;
      menuroot.checkItem( "View", VIEW.LTL2, false) ;
      menuroot.checkItem( "View", VIEW.LTL4, false) ;
      menuroot.checkItem( "View", VIEW.LTL8, false) ;
      if( type==0 ){
        switch(size){
        case 1 : menuroot.checkItem( "View", VIEW.BIG1, true) ; break ;
        case 2 : menuroot.checkItem( "View", VIEW.BIG2, true) ; break ;
        case 4 : menuroot.checkItem( "View", VIEW.BIG4, true) ; break ;
        case 8 : menuroot.checkItem( "View", VIEW.BIG8, true) ; break ;
        default: break ;
        }
      }else{
        switch(size){
        case 1 : menuroot.checkItem( "View", VIEW.LTL1, true) ; break ;
        case 2 : menuroot.checkItem( "View", VIEW.LTL2, true) ; break ;
        case 4 : menuroot.checkItem( "View", VIEW.LTL4, true) ; break ;
        case 8 : menuroot.checkItem( "View", VIEW.LTL8, true) ; break ;
        default: break ;
        }
      }
    }
    return ;
  }

  void setARGS(string[] args){
    this.args = args ;
  }

}
